/* -*- c++ -*-

This is Picprog, pic16c84 programmer software for the serial port device.
Copyright  1997 Jaakko Hyvtti

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

The author may be contacted at:

Email: Jaakko.Hyvatti@iki.fi
URL:   http://www.iki.fi/hyvatti/
Phone: +358 40 5011222

Please send any suggestions, bug reports, success stories etc. to the
Email address above.

*/

#include <iostream.h>
#include <iomanip.h>
#include <sysexits.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>

#include "hexfile.h"

static int got_signal = 0;

void
term_handler (int a)
{
  got_signal = 1;
  signal (a, term_handler);
}

int hexfile::load (const char *name)
{
  ifstream f(name);
  char buf [128];
  int line = 0;
  enum formats format = unknown;
  int addr, words, check, sum, word, i;
  int e;

  if (!f) {
    e = errno;
    cerr << name << ":unable to load hexfile:" << strerror (e) << endl;
    return EX_NOINPUT;
  }
  while (f.get (buf, sizeof (buf))) {
    line++;
    char c;
    if (f.get (c) && '\n' != c) {
      cerr << name << ':' << line << ":long input line\n";
      return EX_DATAERR;
    }
    int len = strlen (buf);
    while (len && isspace (buf [len-1]))
      len--;
    buf [len] = '\0';
    if (!strcmp (buf, ":00000001FF")) // eof
      return EX_OK;

    if (!len)
      continue; // empty line, I do not want to whine about it

    i = len - 1;
    while (i && isxdigit (buf [i]))
      i--;

    if (i || ':' != buf [0] || 3 != (len & 3) || len < 15
	|| '0' != buf [7] || '0' != buf [8]) {
      cerr << name << ':' << line << ":invalid input line\n";
      return EX_DATAERR;
    }
    check = strtol (buf + len - 2, 0, 16);
    buf [7] = '\0';
    addr = strtol (buf + 3, 0, 16);
    buf [3] = '\0';
    words = strtol (buf + 1, 0, 16);
    if (!format) {
      if (words * 4 + 11 == len)
	format = ihx16;
      else if (words * 2 + 11 == len)
	format = ihx8m;
      else {
	cerr << name << ':' << line <<
	  ":unknown input format, only ihx8m and ihx16 accepted\n";
	return EX_DATAERR;
      }
    }
    if (words * (ihx16 == format ? 4 : 2) + 11 != len) {
      cerr << name << ':' << line << ":line length mismatch:"
	   << (ihx16 == format ? "ihx16 " : "ihx8m ")
	   << words * (ihx16 == format ? 4 : 2) + 11 << " != " << len << endl;
      return EX_DATAERR;
    }
    sum = words + addr + (addr >> 8);
    if (ihx8m == format) {
      words /= 2;
      addr /= 2;
    }

    if (!((addr >= 0 && addr+words <= 0x400)
	  || (addr >= 0x2000 && addr+words <= 0x2004)
	  || (addr == 0x2007 && words == 1)
	  || (addr >= 0x2100 && addr+words <= 0x2140))) {
      cerr << name << ':' << line << ":invalid address " << hex
	   << setw(4) << setfill('0') << addr << dec
	   << ", possibly not a pic16c84 hex file?\n";
      return EX_DATAERR;
    }

    while (words--) {
      buf [13 + words * 4] = '\0';
      word = strtol (buf + 9 + words * 4, 0, 16);
      if (ihx8m == format)
	word = (word & 0xff) << 8 | word >> 8;
      sum += word + (word >> 8);
      pgm [addr + words] = word;
    }
    if (sum + check & 0xff) {
      cerr << name << ':' << line << ":checksum mismatch\n";
      return EX_DATAERR;
    }
  }
  e = errno;
  if (!f.eof ()) {
    cerr << name << ':' << line << ':' << strerror (e) << ":\n";
    return EX_IOERR;
  }
  cerr << name << ':' << line << ":warning:unexpected eof\n";
  return EX_OK;
}

void hexfile::save_line (ofstream& f, int begin, int len, enum hexfile::formats format) const
{
  int p_begin, p_len, sum, i;

  if (ihx8m == format) {
    p_begin = begin * 2;
    p_len = len * 2;
  } else {
    p_begin = begin;
    p_len = len;
  }

  f << ':' << setw (2) << p_len << setw (4) << p_begin << "00";

  sum = p_len + p_begin + (p_begin >> 8);
  for (i = begin; i < begin + len; i++) {
    int word = pgm [i];
    if (ihx8m == format)
      word = (word & 0xff) << 8 | (word & 0xff00) >> 8;
    f << setw (4) << word;
    sum += word + (word >> 8);
  }

  f << setw (2) << (-sum & 0xff) << endl;
}

void hexfile::save_region (ofstream& f, int addr, int end, enum hexfile::formats format, bool skip_ones) const
{
  int len;

  for (;;) {
    while (addr < end
	   && (-1 == pgm [addr]
	       || (skip_ones
		   && (addr >= 0x2100 && addr < 0x2140 ? 0xff : 0x3fff)
		      == pgm [addr])))
      addr++;
    if (addr >= end)
      break;
    len = 1;
    while (len < 8 && addr+len < end
	   && !(-1 == pgm [addr+len]
		|| (skip_ones
		    && (addr+len>= 0x2100 && addr+len < 0x2140 ? 0xff : 0x3fff)
		       == pgm [addr+len])))
      len++;
    save_line (f, addr, len, format);
    addr += len;
  }
}

int hexfile::save (const char *name, enum hexfile::formats format, bool skip_ones) const
{
  ofstream f (name);
  int e;

  if (!f) {
    e = errno;
    cerr << name << ":unable to open save file:" << strerror (e) << endl;
    return EX_IOERR;
  }
  f << hex << setfill ('0') << setiosflags (ios::uppercase);

  save_region (f, 0, 0x400, format, skip_ones);
  save_region (f, 0x2000, 0x2004, format, skip_ones);
  save_region (f, 0x2007, 0x2008, format, skip_ones);
  save_region (f, 0x2100, 0x2140, format, skip_ones);
  f << ":00000001FF\n";
  return EX_OK;
}

/*
It will read both IHX16 and IHX8M (auto-detection), program ID and
fuse locations and data memory.  It verifies location before and after
programming, so unnecessary programming is avoided and errors are
detected.  Code protection is detected before programming, and bulk
erase is executed if other than ID locations are programmed if CP is
enabled.
*/

int hexfile::program_location (picport& pic, int addr, bool data) const
{
  int retval;

  if (-1 != pgm [addr]
      && (retval = pic.command (data ? picport::data_from_data
				: picport::data_from_prog)) != pgm [addr]) {
    if (-1 == retval) {
      cerr << pic.port() << ':' << hex << setfill ('0') << setw (4) << addr
	   << dec << ":unable to read pic while programming\n";
      return EX_IOERR;
    }
    pic.command (data ? picport::data_for_data : picport::data_for_prog,
		 pgm [addr]);
    pic.command (picport::beg_prog);

    cout << hex << setfill ('0') << setw (4) << addr << ' '
	 << setw (4) << pic.address () << ' '
	 << setw (4) << pgm [addr] << dec << endl;
    
    // verify, but do not verify fuses if Code Protect bit is cleared!

    if ((pic.address () != 0x2007 || pgm [addr] & 0x10)
	&& pgm [addr] != pic.command (data ? picport::data_from_data
				   : picport::data_from_prog)) {
      cerr << pic.port() << ':' << hex << setw (4) << setfill ('0') << addr << dec
	   << ":unable to verify pic while programming.\n"
	"Is code protection enabled?  "
	"Use --erase option to disable code protection.\n";
      return EX_IOERR;
    }
  }
  return EX_OK;
}

typedef void (*sig_type)(int);

int hexfile::program (const char *port, bool reset) const
{
  int retval;

  sig_type save_t, save_q, save_i;
  save_t = signal (SIGTERM, term_handler);
  save_q = signal (SIGQUIT, term_handler);
  save_i = signal (SIGINT, term_handler);

  {
    picport pic (port);

    if (reset) {
      pic.reset_code_protection ();
      pic.reset ();
      cout << "Removed code protection.\n";
    }

    cout << "Burning program eeprom,\n";
    while (pic.address () < 0x400) {
      if (EX_OK != (retval = program_location (pic, pic.address (), false)))
	return retval;
      pic.command (picport::inc_addr);
      if (got_signal) {
	cerr << "Exiting.\n";
	return EX_UNAVAILABLE;
      }
    }

    cout << "burning data eeprom,\n";
    while (pic.address () < 0x440) {
      if (EX_OK != (retval = program_location (pic,
					       pic.address () + 0x2100 - 0x400,
					       true)))
	return retval;
      pic.command (picport::inc_addr);
      if (got_signal) {
	cerr << "Exiting.\n";
	return EX_UNAVAILABLE;
      }
    }
    
    cout << "burning id words,\n";
    pic.command (picport::load_conf, 0x3fff); // dummy value
    while (pic.address () < 0x2004) {
      if (EX_OK != (retval = program_location (pic, pic.address (), false)))
	return retval;
      pic.command (picport::inc_addr);
      if (got_signal) {
	cerr << "Exiting.\n";
	return EX_UNAVAILABLE;
      }
    }

    pic.command (picport::inc_addr);
    pic.command (picport::inc_addr);
    pic.command (picport::inc_addr);
    
    cout << "burning fuses,\n";
    if (EX_OK != (retval = program_location (pic, pic.address (), false)))
      return retval;
    cout << "done.\n";
  }  // scope of pic

  signal (SIGTERM, save_t);
  signal (SIGQUIT, save_q);
  signal (SIGINT, save_i);
  if (got_signal) {
    cerr << "Exiting.\n";
    return EX_UNAVAILABLE;
  }

  return EX_OK;
}


int hexfile::read (const char *port)
{

  sig_type save_t, save_q, save_i;
  save_t = signal (SIGTERM, term_handler);
  save_q = signal (SIGQUIT, term_handler);
  save_i = signal (SIGINT, term_handler);

  {
    picport pic (port);

    cout << "Reading program eeprom,\n";
    while (pic.address () < 0x400) {
      if (-1 == (pgm [pic.address ()] = pic.command (picport::data_from_prog))) {
	cerr << port << ':' << hex << setfill ('0') << setw (4) << pic.address () << dec
	     << ":unable to read pic\n";
	return EX_IOERR;
      }
      pic.command (picport::inc_addr);
      if (got_signal) {
	cerr << "Exiting.\n";
	return EX_UNAVAILABLE;
      }
    }

    cout << "reading data eeprom,\n";
    while (pic.address () < 0x440) {
      if (-1 == (pgm [pic.address () + 0x2100-0x400]
		 = pic.command (picport::data_from_data))) {
	cerr << port << ':' << hex << setfill ('0') << setw (4)
	     << pic.address () + 0x2100-0x400 << dec
	     << ":unable to read pic data memory\n";
	return EX_IOERR;
      }
      pic.command (picport::inc_addr);
      if (got_signal) {
	cerr << "Exiting.\n";
	return EX_UNAVAILABLE;
      }
    }
    pic.command (picport::load_conf, 0);

    cout << "reading id words,\n";
    while (pic.address () < 0x2004) {
      if (-1 == (pgm [pic.address ()] = pic.command (picport::data_from_prog))) {
	cerr << port << ':' << hex << setfill ('0') << setw (4) << pic.address () << dec
	     << ":unable to read pic id words\n";
	return EX_IOERR;
      }
      pic.command (picport::inc_addr);
      if (got_signal) {
	cerr << "Exiting.\n";
	return EX_UNAVAILABLE;
      }
    }

    pic.command (picport::inc_addr);
    pic.command (picport::inc_addr);
    pic.command (picport::inc_addr);

    // fuses
    cout << "reading fuses,\n";
    if (-1 == (pgm [pic.address ()] = pic.command (picport::data_from_prog))) {
      cerr << port << ':' << hex << setfill ('0') << setw (4) << pic.address () << dec
	   << ":unable to read pic fuses\n";
      return EX_IOERR;
    }
    cout << "done.\n";
  } // pic scope

  signal (SIGTERM, save_t);
  signal (SIGQUIT, save_q);
  signal (SIGINT, save_i);
  if (got_signal) {
    cerr << "Exiting.\n";
    return EX_UNAVAILABLE;
  }

  return EX_OK;
}

