/* -*- c++ -*-

This is Picprog, pic16c84 programmer software for the serial port device.
Copyright  1997 Jaakko Hyvtti

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

The author may be contacted at:

Email: Jaakko.Hyvatti@iki.fi
URL:   http://www.iki.fi/hyvatti/
Phone: +358 40 5011222

Please send any suggestions, bug reports, success stories etc. to the
Email address above.

*/

#include <iostream.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <sysexits.h>
#include <string.h>

#include "picport.h"

const int picport::dtr_bit = TIOCM_DTR;
const int picport::rts_bit = TIOCM_RTS;

picport::picport (const char *tty)
  : addr (0), debug_on (0)
{
  if (0 > (fd = open (tty, O_RDWR))) {
    int e = errno;
    cerr << "Unable to open tty " << tty << ":" << strerror (e) << "\n";
    exit (EX_IOERR);
  }
  tcgetattr (fd, &saved);
  termstate = saved;
  termstate.c_iflag = IGNBRK | IGNPAR;
  termstate.c_oflag = 0;
  termstate.c_cflag = B9600 | CS8 | CREAD | CLOCAL;
  termstate.c_lflag = 0;
  tcsetattr (fd, TCSANOW, &termstate);

  // Initialize lines for programming

  // RTS/DTR low, sleep and then TxD high (program voltage)

  ioctl (fd, TIOCCBRK, 0);
  if (0 > ioctl (fd, TIOCMBIC, &picport::rts_bit)
      || 0 > ioctl (fd, TIOCMBIC, &picport::dtr_bit)) {
    int e = errno;
    tcsetattr (fd, TCSANOW, &saved);
    cerr << "Unable to clear RTS/DTR on tty " << tty << ":" << strerror (e) << "\n";
    exit (EX_IOERR);
  }
  usleep (50000);
  if (0 > ioctl (fd, TIOCSBRK, 0)) {
    int e = errno;
    ioctl (fd, TIOCCBRK, 0);
    tcsetattr (fd, TCSANOW, &saved);
    cerr << "Unable to start break on tty " << tty << ":" << strerror (e) << "\n";
    exit (EX_IOERR);
  }
  usleep (1);
  portname = new char [strlen (tty) + 1];
  strcpy (portname, tty);
}

picport::~picport ()
{
  ioctl (fd, TIOCCBRK, 0);
  usleep (1);
  tcsetattr (fd, TCSANOW, &saved);
  close (fd);
  delete [] portname;
}

void picport::reset ()
{
  ioctl (fd, TIOCMBIC, &picport::rts_bit);
  ioctl (fd, TIOCMBIC, &picport::dtr_bit);
  ioctl (fd, TIOCCBRK, 0);
  usleep (50000);
  ioctl (fd, TIOCSBRK, 0);
  usleep (1);
  addr = 0;
}

void picport::reset_code_protection ()
{
  int i;

  command (load_conf, 0x3fff);
  for (i = 0; i < 7; i++)
    command (inc_addr);
  command (command1);
  command (command7);
  command (beg_prog);
  command (command1);
  command (command7);
}

// for debugging purposes: put data bit up
void picport::force ()
{
  ioctl (fd, TIOCMBIS, &picport::dtr_bit);
}

static void delay_100 ()
{
#ifdef NANOSLEEP_WORKS
  static const struct timespec ns100 = { 0, 100 };
  nanosleep (&ns100, 0);
#else
#ifdef USLEEP_WORKS
  usleep (1);
#else
  volatile int i;
  for (i = 0; i < 10; i++)
    ;
#endif
#endif
}

static void delay_1000 ()
{
#ifdef NANOSLEEP_WORKS
  static const struct timespec ns1000 = { 0, 1000 };
  nanosleep (&ns1000, 0);
#else
#ifdef USLEEP_WORKS
  usleep (1);
#else
  volatile int i;
  for (i = 0; i < 100; i++)
    ;
#endif
#endif
}

void picport::p_out (int b)
{
  ioctl (fd, b ? TIOCMBIS : TIOCMBIC, &picport::dtr_bit); // set data
  ioctl (fd, TIOCMBIS, &picport::rts_bit); // clock up
  delay_100 ();
  ioctl (fd, TIOCMBIC, &picport::rts_bit); // clock down
  delay_100 ();
}

int picport::p_in ()
{
  ioctl (fd, TIOCMBIS, &picport::dtr_bit); // set data up
  ioctl (fd, TIOCMBIS, &picport::rts_bit); // clock up
  delay_100 ();
  ioctl (fd, TIOCMBIC, &picport::rts_bit); // clock down
  delay_100 ();

  int i;

  ioctl (fd, TIOCMGET, &i);
  return 0 != (i & TIOCM_CTS);
}

// -1 == error, no programmer present

int picport::command (enum commands comm, int data = 0)
{
  int tmp1, tmp2;

  // first, send out the command, 6 bits

  int i, shift = comm;
  for (i = 0; i < 6; i++)
    p_out ((shift >> i) & 1);

  shift = 0; // default return value

  switch (comm) {
  case beg_prog:
    usleep (10000);
    break;

  case inc_addr:
    if (++addr >= 0x4000)
      addr = 0x2000;
    break;

  case data_from_prog:
  case data_from_data:
    delay_1000 ();
    tmp1 = p_in ();
    for (i = 0; i < 14; i++)
      shift |= p_in () << i;
    tmp2 = p_in ();

    // Check that the leftover bits were valid, all 1's.
    // This detects if the programmer is not connected to the port.

    if (!tmp1 || !tmp2) {
      cerr << portname << ":PIC programmer or chip fault\n";
      return -1;
    }
    if (data_from_data == comm && (shift & 0x3f00) != 0x3f00) {
      cerr << portname
	   << ":PIC programmer or chip fault\n"
	"Is code protection enabled?  "
	"Use --erase option to disable code protection.\n";
      return -1;
    }

    if (data_from_data == comm)
      shift &= 0xff;
    break;

  case load_conf:
    addr = 0x2000;
    // FALLTHROUGH

  case data_for_prog:
  case data_for_data:
    delay_1000 ();
    p_out (0);
    for (i = 0; i < 14; i++)
      p_out ((data >> i) & 1);
    p_out (0);
    break;

  default:
    ;
  }

  delay_1000 ();
  return shift;
}

