/* -*- c++ -*-

This is a generic string class v1.0.
Copyright  1997 Jaakko Hyvtti

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

The author may be contacted at:

Email: Jaakko.Hyvatti@iki.fi
URL:   http://www.iki.fi/hyvatti/
Phone: +358 40 5011222

Please send any suggestions, bug reports, success stories etc. to the
Email address above.

*/

#ifndef H_STRINGC
#define H_STRINGC

#include <iostream.h>
#include <string.h>
#include <sys/types.h>

class string {

  // Undefined string is indicated by !s .
  // I like unsigned characters.  They can contain ISO-8859-1.
  unsigned char *s;
  size_t len, allocated;

public:

  // Default constructor creates an undefined string.
  string () { s = 0; }

  // Constructor with size parameter creates a defined 0-length string.
  string (size_t initlen) {
    if (initlen) {
      s = new unsigned char [allocated = initlen + 1];
      len = 0;
      s [0] = '\0';
    } else
      s = 0;
  }

  // 'char *' constructors always create a defined string.
  string (const unsigned char *s1) {
    s = new unsigned char [allocated = (len = strlen ((const char *)s1))+1];
    memcpy (s, s1, len+1);
  }
  string (const char *s1) {
    s = new unsigned char [allocated = (len = strlen (s1))+1];
    memcpy (s, s1, len+1);
  }
  string (const unsigned char *s1, size_t len1) {
    s = new unsigned char [allocated = (len = len1) + 1];
    memcpy (s, s1, len);
    s [len] = '\0';
  }
  string (const char *s1, size_t len1) {
    s = new unsigned char [allocated = (len = len1) + 1];
    memcpy (s, s1, len);
    s [len] = '\0';
  }

  // Copy constructor copies the undefinedness or definedness.
  string (const string& s1) {
    if (s1.s) {
      s = new unsigned char [allocated = (len = s1.len) + 1];
      memcpy (s, s1.s, s1.len);
      s [len] = '\0';
    } else
      s = 0;
  }

  // Self-explanatory.
  ~string () {
    if (s)
      delete[] s;
  }

  // Self-explanatory.
  string& operator= (const unsigned char *s1);
  string& operator= (const char *s1) {
    return operator= ((const unsigned char *)s1);
  }
  string& operator= (const string& s1);
  string& operator+= (const string& s1);

  // Creates a one character string.
  string& operator= (int c);
  // Appends one character.
  string& operator+= (int c);

  // Returns a reference to a character in the string.  You may even modify it.
  unsigned char& operator[] (size_t i) {
    if (!s)
      return s [0]; // makes the caller core dump, hahaa!
    if (i >= len)
      return s [len];
    return s [i];
  }

  // Same, but for unmodifiable const strings (preserves constness).
  const unsigned char& operator[] (size_t i) const {
    if (!s)
      return s [0]; // makes the caller core dump, hahaa!
    if (i >= len)
      return s [len];
    return s [i];
  }

  // Extends allocation to at least newlen, returns new max string size.
  size_t allocate (size_t newlen);

  // Returns max string size.
  size_t allocate () const {
    if (s)
      return allocated-1;
    return 0;
  }

  // Returns the null terminated C string pointer.
  // Use like: write (1, str.asciz (), str.length ());
  unsigned char *asciz () const { return s; }

  // Current string content length.
  size_t length () const {
    if (s)
      return len;
    return 0;
  }

  // Uh.. never mind.
  void debug () const;

  // Tells if the string is defined or undefined, handy in: if (str) { }
  operator bool () const {
    return 0 != s;
  }

  // Makes the string undefined, frees it's contents.
  void free () {
    if (s) {
      delete[] s;
      s = 0;
    }
  }
};

// Self-explanatory.
string operator+ (const string& s1, const string& s2);

// Outputs a string, including any null characters.
extern ostream& operator<< (ostream& o, const string& z);

// Inputs a string.
extern istream& operator>> (istream& i, string& z);

// Self-explanatory.
inline bool operator== (const string& s1, const string& s2)
{
  if (!s1 || !s2)	// If either one is undefined..
    return !s1 && !s2;	// ..equal only if both undefined.

  return s1.length () == s2.length ()
    && !memcmp (s1.asciz (), s2.asciz (), s1.length ());
}

inline bool operator!= (const string& s1, const string& s2)
{
  return !operator== (s1, s2);
}

bool operator< (const string& s1, const string& s2);

inline bool operator<= (const string& s1, const string& s2)
{
  return !operator< (s2, s1);
}

inline bool operator> (const string& s1, const string& s2)
{
  return operator< (s2, s1);
}

inline bool operator>= (const string& s1, const string& s2)
{
  return !operator< (s1, s2);
}


#endif // H_STRINGC
